"""
Constants which define ranges of NHS Numbers

License: MIT (https://www.opensource.org/licenses/mit-license.php)

Contributors:
* Marcus Baw <marcusbaw@gmail.com>
* Andy Law <andy.law@roslin.ed.ac.uk>

The most accurate information on the NHS Number ranges we could find is
collated here.
This file should be considered the source of truth for NHS Number ranges.
Ranges are INCLUSIVE of the start and end values.

    0000000010 - 0099999999 Unreserved (not in use)
    0100000000 - 3112999999 Scotland CHI numbers (DDMMYYxxxn)
    3113000000 - 3199999999 Unreserved
    3200000000 - 3999999999 Northern Ireland H&C Numbers
    4000000000 - 4999999999 England, Wales and IOM NHS Numbers
    5000000000 - 5999999999 Not to be issued
    6000000000 - 7999999999 England, Wales and IOM NHS Numbers
    8000000000 - 8599999999 Used within the Republic of Ireland Individual
                            Health Identifier (IHI)
    8600000000 - 8999999999 Unreserved
    9000000000 - 9999999999 Not to be issued (Synthetic/test patients PDS)
"""


# using dataclasses would force Python 3.7 or above so we'll use a simple
# class instead
class Range:
    """
    A class to represent contiguous blocks of allocated NHS Numbers
    """

    start: int
    end: int

    def __init__(self, start, end, label):
        self.start = start
        self.end = end
        self.label = label

    def contains_number(self, number):
        if self.start <= int(number) <= self.end:
            return True
        else:
            return False


class Region:
    """
    A class to represent the blocks of numbers allocated to a given region
    """

    ranges: list
    label: str
    tags: list

    def __init__(self, label, tags, ranges):
        self.label = label
        self.tags = tags
        self.ranges = ranges

    def contains_number(self, number):
        for _range in self.ranges:
            if _range.contains_number(number):
                return True
        return False


RANGE_UNALLOCATED_1 = Range(
    start=10,
    end=9_999_999,
    label="Unallocated 1 (not in use)",
)

RANGE_SCOTLAND = Range(
    start=10_000_000,
    end=3_112_999_999,
    label="Scotland CHI numbers",
)

RANGE_UNALLOCATED_2 = Range(
    start=3_113_000_000,
    end=3_199_999_999,
    label="Unallocated 2 (not in use)",
)

RANGE_NORTHERN_IRELAND = Range(
    start=3_200_000_000,
    end=3_999_999_999,
    label="Northern Ireland H&C Numbers",
)

RANGE_ENGLAND_WALES_IOM_1 = Range(
    start=4_000_000_000,
    end=4_999_999_999,
    label="England Wales and IOM NHS Numbers Range 1",
)

RANGE_RESERVED = Range(
    start=5_000_000_000,
    end=5_999_999_999,
    label="Reserved Range - not to be issued",
)

RANGE_ENGLAND_WALES_IOM_2 = Range(
    start=6_000_000_000,
    end=7_999_999_999,
    label="England Wales and IOM NHS Numbers Range 2",
)

RANGE_EIRE = Range(
    start=8_000_000_000,
    end=8_599_999_999,
    label=(
        "Used within the Republic of Ireland Individual Health Identifier"
        " (IHI)"
    ),
)

RANGE_UNALLOCATED_3 = Range(
    start=8_600_000_000,
    end=8_999_999_999,
    label="Unallocated 3 (not in use)",
)

RANGE_NOT_ISSUED_SYNTHETIC = Range(
    start=9_000_000_000,
    end=9_999_999_999,
    label="Not to be issued (Synthetic/test patients PDS)",
)

FULL_RANGE = Range(
    start=10,
    end=9_999_999_999,
    label="Full range of possible numbers, not all are actually valid",
)


REGION_SCOTLAND = Region(
    label="Scotland CHI numbers",
    tags=["scotland", "chi"],
    ranges=[RANGE_SCOTLAND],
)

REGION_ENGLAND_WALES_IOM = Region(
    label="England Wales and IOM NHS Numbers",
    tags=["england-wales", "e-w-iom", "isle-of-man", "cymru", "wales"],
    ranges=[RANGE_ENGLAND_WALES_IOM_1, RANGE_ENGLAND_WALES_IOM_2],
)

REGION_ENGLAND = REGION_WALES = REGION_IOM = REGION_ENGLAND_WALES_IOM

REGION_NORTHERN_IRELAND = Region(
    label="Northern Ireland H&C Numbers",
    tags=["northern-ireland", "ni", "tuaisceart-éireann"],
    ranges=[RANGE_NORTHERN_IRELAND],
)

REGION_EIRE = Region(
    label=(
        "Used within the Republic of Ireland Individual Health Identifier"
        " (IHI)"
    ),
    tags=[
        "eire",
        "republic-of-ireland",
        "ihi",
        "individual-health-identifier",
        "poblacht-na-héireann",
    ],
    ranges=[RANGE_EIRE],
)

REGION_SYNTHETIC = Region(
    label="Not to be issued (Synthetic/test patients PDS)",
    tags=["test", "synthetic"],
    ranges=[RANGE_NOT_ISSUED_SYNTHETIC],
)

REGION_UNALLOCATED = Region(
    label="Unallocated - should not be a valid Number",
    tags=["unallocated"],
    ranges=[RANGE_UNALLOCATED_1, RANGE_UNALLOCATED_2, RANGE_UNALLOCATED_3],
)

REGION_RESERVED = Region(
    label="Reserved and not issued", tags=["reserved"], ranges=[RANGE_RESERVED]
)

REGIONS = {
    "UNALLOCATED": REGION_UNALLOCATED,
    "SCOTLAND": REGION_SCOTLAND,
    "NORTHERN_IRELAND": REGION_NORTHERN_IRELAND,
    "ENGLAND_WALES_IOM": REGION_ENGLAND_WALES_IOM,
    "RESERVED": REGION_RESERVED,
    "EIRE": REGION_EIRE,
    "SYNTHETIC": REGION_SYNTHETIC,
}



"""
Generates valid or invalid NHS numbers, for testing

License: MIT (https://www.opensource.org/licenses/mit-license.php)

Contributors
* Marcus Baw <marcusbaw@gmail.com>
"""
import random

# standard imports
from random import randint

# third-party imports

# local imports
from nhs_number.validate import calculate_checksum
from nhs_number import Region, FULL_RANGE


def generate(
    valid: bool = True,
    for_region: Region = None,
    quantity: int = 1,
) -> list:

    """
    Generates valid or invalid NHS numbers, for testing.

    IMPORTANT: This function is not intended to be used to create real NHS
    numbers. It is for testing purposes only.
    All live NHS numbers are allocated centrally and are not generated locally.

    Optional keyword arguments:

        valid: bool
            Determines whether the NHS numbers to be generated are valid or
            invalid.
        for_region: Region
            If supplied, the start and end ranges for the supplied region will
            be used and any start and end ranges supplied in the function
            call will be ignored.
        quantity: int
            How many NHS numbers to generate.

    The algorithm used here is very naive and could be significantly optimised.
    But it is fairly fast anyway and works well enough for the purposes of
    this library.
    Contributions are welcome.
    """

    # initialise the list of NHS numbers to be returned
    nhs_numbers = []

    ranges = [FULL_RANGE]

    if for_region:
        if isinstance(for_region, Region):
            ranges = for_region.ranges
        else:
            raise TypeError("The for_region argument must be of type Region")

    # generates valid random NHS numbers within the given REGION range
    while len(nhs_numbers) < quantity:
        this_range = random.choice(ranges)
        candidate = randint(this_range.start, this_range.end) // 10
        candidate_str = str(candidate).zfill(9)
        checksum_str = str(calculate_checksum(candidate_str))
        if len(checksum_str) == 1:
            if valid:
                nhs_numbers.append(candidate_str + checksum_str)
            else:
                wrong_checksum_str = checksum_str
                while wrong_checksum_str == checksum_str:
                    wrong_checksum_str = str(random.choice(range(0, 10)))
                nhs_numbers.append(candidate_str + wrong_checksum_str)

    return nhs_numbers



"""
Returns more detailed information on NHS numbers.

License: MIT (https://www.opensource.org/licenses/mit-license.php)

Contributors
* Marcus Baw <marcusbaw@gmail.com>
"""

# standard imports

# third-party imports

# local imports
from nhs_number import REGIONS
from nhs_number.validate import is_valid, calculate_checksum


class NhsNumber:
    nhs_number: str
    """
    A class which returns more information about an NHS Number than simply
    Boolean validity.

    nhs_number: str
        The NHS number to be validated, returned to you as a sense-check.
    identifier_digits: str
        The first 9 digits of the NHS number, which are the identifier digits.
    check_digit: int
        The 10th digit of the NHS number, which is the check digit.
    valid: bool
        Whether the NHS number is valid or not according to the checksum
        comparison.
    calculated_checksum: int
        The checksum calculated from the identifier digits, so you can compare
        it to the check digit.
    region: Region
        The region the NHS number is valid in, if any. Returns an instance of
        the Region class from the constants module.
    region_comment: str
        The name of the region the NHS number is valid in, if any. Returns a
        string.

    Usage:
    >>> from nhs_number import NhsNumber
    >>> nhs_number = NhsNumber('9876543210')
    >>> vars(nhs_number)
    {'nhs_number': '9876543210', 'identifier_digits': '987654321', 'check_digit': 0, 'valid': True, 'calculated_checksum': 0, 'region': <nhs_number.constants.Region object at 0x7fcb31de5e90>, 'region_comment': 'Not to be issued (Synthetic/test patients PDS)'}

    """

    def __init__(self, nhs_number):
        self.nhs_number = nhs_number

        # split the number into its parts
        (identifier_digits, check_digit) = (
            nhs_number[:-1],
            int(nhs_number[-1]),
        )
        self.identifier_digits = identifier_digits
        self.check_digit = check_digit

        # determine if the number is valid
        valid = is_valid(nhs_number)
        self.valid = valid

        # determine if reason for invalidity is failure of checksum match
        calculated_checksum = calculate_checksum(identifier_digits)
        self.calculated_checksum = calculated_checksum

        # default region comment is 'not found'
        region_comment = "Number did not match a known NHS number range"

        # determine the region the number is valid in
        for handle, region in REGIONS.items():
            if region.contains_number(nhs_number):
                region_comment = region.label
                self.region = region

        self.region_comment = region_comment



"""
Validates NHS Numbers

Originally derived from Andy Law's nhs_number Python package validation code

License: MIT (https://www.opensource.org/licenses/mit-license.php)

Contributors
* Andy Law <andy.law@roslin.ed.ac.uk>
* Marcus Baw <marcusbaw@gmail.com>
"""
from __future__ import annotations # for Python 3.7 (remove once we stop supporting 3.7)
from nhs_number.standardise import standardise_format
from nhs_number.constants import Region


def calculate_checksum(identifier_digits: str) -> int | None:
    """
    Calculates the checksum digit for the supplied NHS number (as a string) and
    returns it as an integer.
    IMPORTANT: Expects a string of NINE digits (ie REMOVE the check digit,
    don't supply a full NHS number)
    """
    if len(identifier_digits) != 9:
        return None

    # For the algorithm that calculates the checksum digit, see
    # https://www.closer.ac.uk/wp-content/uploads/CLOSER-NHS-ID-Resource-Report-Apr2018.pdf
    parts_list = [
        int(digit) * (10 - index)
        for index, digit in enumerate(identifier_digits)
    ]
    list_sum = sum(parts_list)
    checksum = 11 - (list_sum % 11)
    if checksum == 11:
        checksum = 0
    return checksum


def is_valid(nhs_number: str, for_region: Region = None) -> bool:
    """
    Checks the supplied NHS number (as a string) is valid and returns True
    or False.

    Internally this uses the normalise_number() function to check that this is
    a valid format to start with.
    Any non-valid input will result in returning False.

    The NHS number must:
        1. be 10 digits long
        2. The 10th digit is a check digit to confirm validity using the
           modulus 11 method

    How NHS Number validation works:
            https://www.datadictionary.nhs.uk/attributes/nhs_number.html

    """
    # Normalise the NHS number to remove any spaces or dashes
    nhs_number = standardise_format(nhs_number)
    if not nhs_number:
        return False

    # If the NHS number is outside the range for the supplied region,
    # return False
    if for_region:
        if not for_region.contains_number(nhs_number):
            return False
        # Additional checks for Scotland CHI number DOB validity will go here

    # Test for checksum validity
    # The first 9 numbers are used to calculate the checksum, which should
    # match the last digit
    (identifier_digits, check_digit) = (nhs_number[:-1], int(nhs_number[-1]))
    calculated_checksum = calculate_checksum(identifier_digits)
    # NOTE: a checksum of 10 is invalid (and is quoted as a special case),
    # but the check for equality
    # will catch that circumstance anyway
    return False if calculated_checksum != check_digit else True
